from __future__ import annotations

from typing import Self, TypeVar, override

import numpy as np
from jax import tree

from .annotations import JaxArray, NumpyRealArray

T = TypeVar('T')


def project_tree(projectable: T, /, *, projector: Projector) -> T:
    """Project the matrices in a PyTree to the plane.

    This is useful for graphing.
    """
    return tree.map(projector.project, projectable, is_leaf=lambda x: isinstance(x, Projectable))


class Projectable:
    """Base class to override how an element of a PyTree is projected."""
    def project(self, projector: Projector) -> Self:
        raise NotImplementedError


class Projector:
    @override
    def __init__(self, *, seed: int = 0, dimensions: int = 2) -> None:
        super().__init__()
        self.seed = seed
        self.dimensions = dimensions
        self.projection_matrices: dict[int, NumpyRealArray] = {}

    def project(self, projectable: T) -> T:
        if projectable is None:
            return projectable
        if isinstance(projectable, Projectable):
            return projectable.project(self)

        assert isinstance(projectable, (JaxArray, np.ndarray))
        features = projectable.shape[-1]
        if features <= self.dimensions:
            return projectable
        projection_matrix = self.get_projection_matrix(features)
        return projectable @ projection_matrix  # type: ignore[return-value] # pyright: ignore

    def get_projection_matrix(self, features: int) -> NumpyRealArray:
        if features not in self.projection_matrices:
            self.projection_matrices[features] = _random_directions(self.seed, features,
                                                                    self.dimensions)
        return self.projection_matrices[features]


def _random_directions(seed: int, dimensions: int, n_axes: int) -> NumpyRealArray:
    """Produce a random projection matrix.

    Returns: a matrix of random numbers with `n_axes` columns, of length `dimensions`, each of
    which is mutually orthogonal and unit-length.
    """
    if dimensions < n_axes:
        raise ValueError
    if seed == 0 and dimensions == n_axes:
        return np.eye(dimensions)
    np_rng = np.random.default_rng(seed)
    a = np_rng.uniform(size=(dimensions, n_axes))
    directions, _ = np.linalg.qr(a, mode='reduced')
    assert isinstance(directions, np.ndarray)
    return directions



import jax.numpy as jnp
import pytest
from jax import Array, vmap

try:
    from flax import nnx

    from tjax.dataclasses import DataClassModule
except ImportError:
    pytest.skip("Skipping NNX graph test", allow_module_level=True)


def test_dataclass_module() -> None:
    class SomeModule(nnx.Module):
        def __init__(self, epsilon: Array) -> None:
            super().__init__()
            self.epsilon = nnx.Variable(epsilon)

    class SomeDataclassModule(DataClassModule):
        def __init__(self, rngs: nnx.Rngs) -> None:
            super().__init__(rngs=rngs)
            self.sm = SomeModule(jnp.zeros(1))

    def f(m: SomeDataclassModule, x: Array) -> None:
        pass

    rngs = nnx.Rngs()
    module = SomeDataclassModule(rngs)
    z = jnp.zeros(10)
    vmap(f, in_axes=(None, 0))(module, z)



from __future__ import annotations

from collections.abc import Callable, Hashable, Iterable
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from jax.tree_util import register_pytree_with_keys

from ..annotations import PyTree
from .types_ import GraphAuxData, GraphData, GraphEdgeKey, GraphNodeKey, UndirectedGraphEdgeKey

try:
    import networkx as nx
except ImportError:
    msg = "NetworkX not available"
    def register_graph_as_jax_pytree(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
    def register_graph_as_nnx_node(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
else:
    if TYPE_CHECKING:
        Graph: TypeAlias = nx.Graph[Any]
    else:
        Graph: TypeAlias = nx.Graph

    def nodes_helper(graph: nx.Graph[Any]) -> tuple[list[GraphNodeKey], list[PyTree]]:
        node_data = graph.nodes.data()
        node_keys = [GraphNodeKey(name) for name, _ in node_data]
        node_values = [value for _, value in node_data]
        return node_keys, node_values

    def edges_helper(graph: nx.Graph[Any]) -> tuple[list[GraphEdgeKey], list[PyTree]]:
        directed = isinstance(graph, nx.DiGraph)
        edge_data = graph.edges.data()
        edge_key: Callable[[Hashable, Hashable], GraphEdgeKey] = (
                GraphEdgeKey if directed else UndirectedGraphEdgeKey)
        edge_keys = [edge_key(source, target) for source, target, _ in edge_data]
        edge_data = [data for _, _, data in edge_data]
        return edge_keys, edge_data

    def flatten_tree(graph: nx.Graph[Any], /) -> tuple[list[GraphData], GraphAuxData]:
        node_keys, node_values = nodes_helper(graph)
        edge_keys, edge_values = edges_helper(graph)
        return node_values + edge_values, node_keys + edge_keys

    def register_graph_as_jax_pytree(graph_type: type[nx.Graph[Any]]) -> None:  # pyright: ignore
        def unflatten_tree(keys: GraphAuxData, values: Iterable[GraphData], /) -> nx.Graph[Any]:
            graph = graph_type()
            for key, value in zip(keys, values, strict=True):
                match key:
                    case GraphNodeKey(name):
                        graph.add_node(name, **value)
                    case GraphEdgeKey(source, target):
                        graph.add_edge(source, target, **value)
            return graph

        def flatten_with_keys(graph: nx.Graph[Any], /
                              ) -> tuple[Iterable[tuple[GraphNodeKey | GraphEdgeKey, GraphData]],
                                         GraphAuxData]:
            values, keys = flatten_tree(graph)
            return (zip(keys, values, strict=True), keys)

        flatten_with_keys_ = cast(
                "Callable[[Graph], tuple[Iterable[tuple[Hashable, Any]], Hashable]]",
                flatten_with_keys)
        unflatten_tree_ = cast("Callable[[Hashable, Any], Graph]", unflatten_tree)
        flatten_tree_ = cast("Callable[[Graph], tuple[Iterable[Any], Hashable]]",
                             flatten_tree)
        register_pytree_with_keys(graph_type, flatten_with_keys_, unflatten_tree_, flatten_tree_)



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp
import jax.random as jr
import numpy as np
from jax import Array, debug, tree
from rich.console import Console

from ..dataclasses.dataclass import dataclass
from .internal import internal_print_generic


@dataclass
class JaxKey:
    key_data: Array


def replace_key(leaf: Any) -> Any:
    if isinstance(leaf, Array) and jnp.issubdtype(leaf.dtype, jax.dtypes.prng_key):
        return JaxKey(jr.key_data(leaf))
    return leaf


def print_generic(*args: object,
                  immediate: bool = False,
                  raise_on_nan: bool = True,
                  as_leaves: bool = False,
                  console: Console | None = None,
                  ) -> None:
    """Uses internal_print_generic in a tapped function.

    Args:
        immediate: Print the value immediately even if they're tracers.  Do not wait for the
            compiled function to be called.
        raise_on_nan: Assert if NaN is found in the output.
        as_leaves: Print the elements of args showing only the the tree leaves and their paths.
        console: The console that formats the output.
        args: Positional arguments to be printed.  Only dynamic arguments are allowed.
    """
    args = tree.map(replace_key, args)

    if immediate:
        internal_print_generic(*args, raise_on_nan=raise_on_nan, as_leaves=as_leaves,
                               console=console)
        return

    leaves, tree_def = tree.flatten(args)

    def fix(u_arg: Any, arg: Any) -> Any:
        return (np.asarray(u_arg) if isinstance(arg, np.ndarray)
                else int(u_arg.item()) if isinstance(arg, int)
                else u_arg)

    def callback(*callback_leaves: object) -> None:
        unflattened_tree = tree.unflatten(tree_def, callback_leaves)
        v_args = tree.map(fix, unflattened_tree, args)
        internal_print_generic(*v_args, raise_on_nan=raise_on_nan, as_leaves=as_leaves,
                               console=console)

    debug.callback(callback, *leaves, ordered=True)



from __future__ import annotations

from collections.abc import Callable
from typing import Generic, TypeVar

import jax.numpy as jnp
from jax import tree
from typing_extensions import override

from tjax.dataclasses import dataclass

from ..annotations import ComplexNumeric, JaxComplexArray, PyTree, RealNumeric
from .transform import GradientState, SecondOrderGradientTransformation

Weights = TypeVar('Weights', bound=PyTree)


@dataclass
class SMDState(GradientState, Generic[Weights]):
    log_learning_rate: Weights
    v: Weights


@dataclass
class SMDGradient(SecondOrderGradientTransformation[SMDState[Weights], Weights],
                  Generic[Weights]):
    """Stochastic meta-descent.

    Schraudolph, N. N. (1999). Local gain adaptation in stochastic gradient descent. Artificial
    Neural Networks, 1999. ICANN 99. Ninth International Conference on (Conf. Publ. No. 470), 2,
    569-574. https://doi.org/10.1049/cp:19991170.
    """
    meta_learning_rate: RealNumeric = 1e-2

    @override
    def init(self, parameters: Weights) -> SMDState[Weights]:
        z = tree.map(jnp.zeros_like, parameters)
        return SMDState[Weights](z, z)

    @override
    def second_order_update(self,
                            gradient: Weights,
                            state: SMDState[Weights],
                            parameters: Weights | None,
                            hessian_vector_product: Callable[[Weights], Weights]
                            ) -> tuple[Weights, SMDState[Weights]]:
        negative_gradient = tree.map(jnp.negative, gradient)  # delta

        # Update log-learning rate.
        def g(log_p: RealNumeric, delta: ComplexNumeric, v: ComplexNumeric) -> JaxComplexArray:
            return jnp.asarray(log_p + self.meta_learning_rate * delta * v)

        new_log_learning_rate = tree.map(g, state.log_learning_rate, negative_gradient, state.v)
        learning_rate = tree.map(jnp.exp, new_log_learning_rate)  # p

        # Calculate gradient.
        gradient = tree.map(jnp.multiply, learning_rate, negative_gradient)

        # Update v.
        def f(v: ComplexNumeric,
              p: RealNumeric,
              delta: ComplexNumeric,
              hv: ComplexNumeric) -> JaxComplexArray:
            return jnp.asarray(v + p * delta - hv)

        new_v = tree.map(f, state.v, learning_rate, negative_gradient,
                         hessian_vector_product(state.v))

        return gradient, SMDState[Weights](new_log_learning_rate, new_v)

# got
# Union[ndarray[Any, dtype[complexfloating[Any, Any]]],
#       ndarray[Any, dtype[floating[Any]]],
#       ndarray[Any, dtype[signedinteger[Any]]],
#       Array,
#       complex]
# , expected
# Union[ndarray[Any, dtype[Union[floating[Any], complexfloating[Any, Any]]]],
#       Array,
#       complex, float, int]



from __future__ import annotations

from typing import Any

import pytest
from jax import jit, tree

from tjax import register_graph_as_jax_pytree

try:
    import networkx as nx
except ImportError:
    pytest.skip("Skipping graph test", allow_module_level=True)
else:
    register_graph_as_jax_pytree(nx.DiGraph)


@pytest.fixture(scope='session')
def graph() -> nx.DiGraph[Any]:
    g = nx.DiGraph()
    g.add_node('a', y=2.0)
    g.add_node('b', z=3.0)
    g.add_node('c', w=4.0)
    g.add_edge('a', 'b', x=5.0)
    g.add_edge('c', 'b', x=7.0)
    return g


@jit
def f(x: nx.DiGraph[Any]) -> nx.DiGraph[Any]:
    return x


def test_rebuild(graph: nx.DiGraph[Any]) -> None:
    values, tree_def = tree.flatten(graph)
    rebuilt_graph = tree.unflatten(tree_def, values)
    assert nx.utils.graphs_equal(graph, rebuilt_graph)


def test_rebuild_jit(graph: nx.DiGraph[Any]) -> None:
    rebuilt_graph = f(graph)
    assert nx.utils.graphs_equal(graph, rebuilt_graph)


def test_flatten_flavors(graph: nx.DiGraph[Any]) -> None:
    values_a, tree_def_a = tree.flatten(graph)
    keys_and_values, tree_def_b = tree.flatten_with_path(graph)
    key_paths, values_b = zip(*keys_and_values, strict=True)
    assert hash(tree_def_a) == hash(tree_def_b)
    assert values_a == list(values_b)
    assert values_a == [2.0, 3.0, 4.0, 5.0, 7.0]
    simplified = {"".join(str(x) for x in key_path)
                  for key_path in key_paths}
    assert simplified == {".node['a']['y']",
                          ".node['b']['z']",
                          ".node['c']['w']",
                          ".edge['a', 'b']['x']",
                          ".edge['c', 'b']['x']"}



from __future__ import annotations

from typing import Any, TypeVar, override

import pytest

from tjax import GraphEdgeKey, register_graph_as_nnx_node

try:
    import networkx as nx
    from flax import nnx
except ImportError:
    pytest.skip("Skipping NNX graph test", allow_module_level=True)
else:
    register_graph_as_nnx_node(nx.DiGraph)


T = TypeVar('T')


class ComparableVariable(nnx.Variable[T]):
    @override
    def __hash__(self) -> int:
        raise NotImplementedError

    @override
    def __eq__(self, other: object) -> bool:
        assert isinstance(other, ComparableVariable)
        return self.raw_value == other.raw_value


@pytest.fixture(scope='session')
def graph() -> nx.DiGraph[Any]:
    v = ComparableVariable(2.0)
    w = ComparableVariable(3.0)
    x = ComparableVariable(4.0)
    g = nx.DiGraph()
    g.add_node('a', y=v)
    g.add_node('b', z=w)
    g.add_edge('a', 'b', x=x)
    g.add_edge('c', 'b', x=x)
    return g


def test_rebuild(graph: nx.DiGraph[Any]) -> None:
    graph_def, state = nnx.graph.flatten(graph)
    rebuilt_graph = nnx.graph.unflatten(graph_def, state)
    assert nx.utils.graphs_equal(graph, rebuilt_graph)


def test_clone(graph: nx.DiGraph[Any]) -> None:
    class Model(nnx.Module):
        def __init__(self) -> None:
            super().__init__()
            self.graph = graph

    m = Model()
    n = nnx.graph.clone(m)
    assert nx.utils.graphs_equal(m.graph, n.graph)


def test_flatten(graph: nx.DiGraph[Any]) -> None:
    _, state = nnx.graph.flatten(graph)
    state_dict = dict(state)
    variable = state_dict[GraphEdgeKey('a', 'b'), 'x']
    assert isinstance(variable, nnx.VariableState)
    assert variable.value == 4.0  # noqa: PLR2004



from io import StringIO

import jax.numpy as jnp
import pytest
from pytest import CaptureFixture
from rich.console import Console

from tjax import print_generic
from tjax.dataclasses import field

from .test_display import verify

try:
    from flax import nnx

    from tjax.dataclasses import DataClassModule
except ImportError:
    pytest.skip("Skipping NNX display tests", allow_module_level=True)


def test_module_display(capsys: CaptureFixture[str],
                        console: Console) -> None:
    class C(nnx.Module):
        def __init__(self) -> None:
            super().__init__()
            self.x = jnp.zeros(4)
            self.y = 'abc'
            self.a = nnx.Variable(jnp.zeros(4))

    c = C()
    print_generic(c, console=console, immediate=True)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           C[flax-module]
           ├── x=Jax Array (4,) float64
           │   └──  0.0000 │ 0.0000 │ 0.0000 │ 0.0000
           ├── y="abc"
           └── a=Variable
               └── value=Jax Array (4,) float64
                   └──  0.0000 │ 0.0000 │ 0.0000 │ 0.0000
           """)


def test_dataclass_module_display(capsys: CaptureFixture[str],
                                  console: Console) -> None:
    class C(DataClassModule):
        x: float
        y: str = field(static=True)

        def __post_init__(self, rngs: nnx.Rngs) -> None:
            if hasattr(super(), '__post_init__'):
                super().__post_init__(rngs)
            self.a = nnx.Variable(jnp.zeros(4))

    c = C(0.5, 'abc', rngs=nnx.Rngs())
    print_generic(c, console=console)
    assert isinstance(console.file, StringIO)
    captured = console.file.getvalue()
    verify(captured,
           """
           C[dataclass,flax-module]
           ├── a=Variable
           │   └── value=Jax Array (4,) float64
           │       └──  0.0000 │ 0.0000 │ 0.0000 │ 0.0000
           ├── x=0.5
           └── y="abc"
           """)



from __future__ import annotations

from collections.abc import Callable, Hashable, Mapping, Sequence
from reprlib import recursive_repr
from typing import Any, Generic, TypeVar, cast

from jax.tree_util import register_pytree_node_class
from typing_extensions import override

from .annotations import PyTree

R = TypeVar('R')


@register_pytree_node_class
class Partial(Generic[R]):
    """A version of functools.partial that returns a pytree.

    Use it for partial function evaluation in a way that is compatible with JAX's transformations,
    e.g., ``Partial(func, *args, **kwargs)``.

    JAX's version of this class in jax.tree_util assumes that all parameters are dynamic and the
    callable is static.  This version allows you to specify which parameters are static and whether
    the callable is static.

    Args:
        function: The function being applied.
        args: The static and dynamic applied positional arguments.
        callable_is_static: Whether the function callable is static.
        static_argnums: The indices of the applied positional arguments that are static.
        static_kwargs: The static applied keyword arguments.
        dynamic_kwargs: The dynamic applied keyword arguments.
    """
    @override
    def __init__(self,
                 function: Partial[R] | Callable[..., R],
                 /,
                 *args: object,
                 callable_is_static: bool = True,
                 static_argnums: tuple[int, ...] = (),
                 static_kwargs: Mapping[str, Any] | None = None,
                 **dynamic_kwargs: object) -> None:
        super().__init__()
        if isinstance(function, Partial) and callable_is_static:
            raise TypeError
            # Could collapse the arguments here, but the benefit is small.
        self.function = function
        self.callable_is_static = callable_is_static
        self.args = args
        self.static_argnums = static_argnums
        self.static_kwargs = {} if static_kwargs is None else static_kwargs
        self.dynamic_kwargs = dynamic_kwargs

    def tree_flatten(self: Partial[R]) -> tuple[Sequence[PyTree], Hashable]:
        static_args, dynamic_args = self._partition_args()
        static_kwargs = tuple((key, self.static_kwargs[key]) for key in sorted(self.static_kwargs))
        return ((dynamic_args, self.dynamic_kwargs),
                (self.callable_is_static, self.static_argnums, static_args, static_kwargs))

    @classmethod
    def tree_unflatten(cls,
                       static: Hashable,
                       trees: Sequence[PyTree]) -> Partial[R]:
        if not isinstance(static, tuple):
            raise RuntimeError  # noqa: TRY004

        callable_is_static, static_argnums, static_args, static_kwarg_items = static
        static_kwargs = dict(static_kwarg_items)

        if not isinstance(static_args, tuple):
            raise RuntimeError  # noqa: TRY004

        dynamic_args, dynamic_kwargs = trees

        if not isinstance(dynamic_args, tuple):
            raise RuntimeError  # noqa: TRY004
        if not isinstance(dynamic_kwargs, dict):
            raise RuntimeError  # noqa: TRY004

        dynamic_kwargs = cast('dict[str, Any]', dynamic_kwargs)
        args = cls._unpartition_args(static_argnums, static_args, dynamic_args,
                                     callable_is_static=callable_is_static)

        return Partial[R](*args,
                          callable_is_static=callable_is_static,
                          static_argnums=static_argnums,
                          static_kwargs=static_kwargs,
                          **dynamic_kwargs)

    # Magic methods --------------------------------------------------------------------------------
    def __call__(self, *args: object, **kwargs: object) -> R:
        keywords = {**self.static_kwargs, **self.dynamic_kwargs, **kwargs}
        return self.function(*self.args, *args, **keywords)

    @recursive_repr()
    def __repr__(self) -> str:
        qualname = type(self).__qualname__
        args = [repr(self.function)]
        args.extend(repr(x) for x in self.args)
        args.extend([f"callable_is_static={self.callable_is_static}",
                     f"static_argnums={self.static_argnums}",
                     f"static_kwargs={self.static_kwargs}"])
        args.extend(f"{k}={v!r}" for (k, v) in self.dynamic_kwargs.items())
        return f"{qualname}({', '.join(args)})"

    # Private methods ------------------------------------------------------------------------------
    def _partition_args(self) -> tuple[tuple[Any, ...], tuple[Any, ...]]:
        # Partition self.args into static and dynamic arguments.
        static_args = []
        dynamic_args = []

        static_argnums = set(self.static_argnums)

        def _append(value: Any, *, is_static: bool) -> None:
            if is_static:
                static_args.append(value)
            else:
                dynamic_args.append(value)

        _append(self.function, is_static=self.callable_is_static)
        for i, value in enumerate(self.args):
            _append(value, is_static=i in static_argnums)

        return tuple(reversed(static_args)), tuple(reversed(dynamic_args))

    @classmethod
    def _unpartition_args(cls,
                          static_argnums: tuple[int, ...],
                          static_args: tuple[Any, ...],
                          dynamic_args: tuple[Any, ...],
                          *,
                          callable_is_static: bool
                          ) -> tuple[Any, ...]:
        static_arg_list = list(static_args)
        dynamic_arg_list = list(dynamic_args)
        args = []
        for i in range(len(static_args) + len(dynamic_args)):
            is_static = (callable_is_static
                         if i == 0
                         else i - 1 in static_argnums)
            if is_static:
                args.append(static_arg_list.pop())
            else:
                args.append(dynamic_arg_list.pop())
        return tuple(args)



"""Tools for the differential programming library [JAX](https://github.com/google/jax)."""
from . import dataclasses, gradient
from ._src.abstract_method_decorators import JaxAbstractClass, abstract_custom_jvp, abstract_jit
from ._src.annotations import (Array, BooleanArray, BooleanNumeric, ComplexArray, ComplexNumeric,
                               IntegralArray, IntegralNumeric, JaxArray, JaxBooleanArray,
                               JaxComplexArray, JaxIntegralArray, JaxRealArray, KeyArray,
                               NumpyArray, NumpyBooleanArray, NumpyBooleanNumeric,
                               NumpyComplexArray, NumpyComplexNumeric, NumpyIntegralArray,
                               NumpyIntegralNumeric, NumpyRealArray, NumpyRealNumeric, PyTree,
                               RealArray, RealNumeric, Shape, ShapeLike, SliceLike)
from ._src.cotangent_tools import (copy_cotangent, cotangent_combinator, print_cotangent,
                                   replace_cotangent, reverse_scale_cotangent, scale_cotangent,
                                   zero_from_primal)
from ._src.display.display_generic import display_generic
from ._src.display.internal import internal_print_generic
from ._src.display.print_generic import print_generic
from ._src.dtype_tools import cast_to_result_type
from ._src.graph.display import graph_arrow, graph_edge_name
from ._src.graph.register_flax import register_graph_as_nnx_node
from ._src.graph.register_jax import register_graph_as_jax_pytree
from ._src.graph.types_ import GraphEdgeKey, GraphNodeKey, UndirectedGraphEdgeKey
from ._src.k_means import cluster_with_k_means
from ._src.leaky_integral import diffused_leaky_integrate, leaky_data_weight, leaky_integrate
from ._src.math_tools import (abs_square, divide_where, inverse_softplus, log_softplus,
                              matrix_dot_product, matrix_vector_mul, normalize, outer_product,
                              softplus, stop_gradient, sublinear_softplus)
from ._src.partial import Partial
from ._src.project_tree import Projectable, Projector, project_tree
from ._src.rng import RngStream, create_streams, fork_streams, sample_streams
from ._src.shims import custom_jvp, custom_jvp_method, custom_vjp, custom_vjp_method, hessian, jit
from ._src.testing import (assert_tree_allclose, get_relative_test_string, get_test_string,
                           tree_allclose)
from ._src.tree_tools import dynamic_tree_all, element_count, tree_sum

__all__ = [
    'Array',
    'BooleanArray',
    'BooleanNumeric',
    'ComplexArray',
    'ComplexNumeric',
    'GraphEdgeKey',
    'GraphNodeKey',
    'IntegralArray',
    'IntegralNumeric',
    'JaxAbstractClass',
    'JaxArray',
    'JaxBooleanArray',
    'JaxComplexArray',
    'JaxIntegralArray',
    'JaxRealArray',
    'KeyArray',
    'NumpyArray',
    'NumpyBooleanArray',
    'NumpyBooleanNumeric',
    'NumpyComplexArray',
    'NumpyComplexNumeric',
    'NumpyIntegralArray',
    'NumpyIntegralNumeric',
    'NumpyRealArray',
    'NumpyRealNumeric',
    'Partial',
    'Projectable',
    'Projector',
    'PyTree',
    'RealArray',
    'RealNumeric',
    'RngStream',
    'Shape',
    'ShapeLike',
    'SliceLike',
    'UndirectedGraphEdgeKey',
    'abs_square',
    'abstract_custom_jvp',
    'abstract_jit',
    'assert_tree_allclose',
    'cast_to_result_type',
    'cluster_with_k_means',
    'copy_cotangent',
    'cotangent_combinator',
    'create_streams',
    'custom_jvp',
    'custom_jvp_method',
    'custom_vjp',
    'custom_vjp_method',
    'dataclasses',
    'diffused_leaky_integrate',
    'display_generic',
    'divide_where',
    'dynamic_tree_all',
    'element_count',
    'fork_streams',
    'get_relative_test_string',
    'get_test_string',
    'gradient',
    'graph_arrow',
    'graph_edge_name',
    'hessian',
    'internal_print_generic',
    'inverse_softplus',
    'jit',
    'leaky_data_weight',
    'leaky_integrate',
    'log_softplus',
    'matrix_dot_product',
    'matrix_vector_mul',
    'normalize',
    'outer_product',
    'print_cotangent',
    'print_generic',
    'project_tree',
    'register_graph_as_jax_pytree',
    'register_graph_as_nnx_node',
    'replace_cotangent',
    'reverse_scale_cotangent',
    'sample_streams',
    'scale_cotangent',
    'softplus',
    'stop_gradient',
    'sublinear_softplus',
    'tree_allclose',
    'tree_sum',
    'zero_from_primal',
]



"""Tests from optax._src.transform_test."""
from collections.abc import Callable
from typing import Any

import chex
import jax.numpy as jnp
import numpy as np
import pytest
from jax import tree
from optax import apply_updates

from tjax import RealArray
from tjax.gradient import (SGD, AddDecayedWeights, ApplyEvery, Centralize,
                           ChainedGradientTransformation, Ema, GradientTransformation, Scale,
                           ScaleByAdam, ScaleByParamBlockNorm, ScaleByParamBlockRMS, ScaleByRms,
                           ScaleByStddev, ScaleByTrustRatio, Trace)

STEPS = 50
LR = 1e-2
init_params = (np.asarray([1., 2.]), np.asarray([3., 4.]))
per_step_updates = (np.asarray([500., 5.]), np.asarray([300., 3.]))


def variant(x: Any) -> Any:
    return x


@pytest.mark.parametrize("constructor", [ScaleByAdam, ScaleByRms, ScaleByStddev, ScaleByTrustRatio,
                                         ScaleByParamBlockNorm, ScaleByParamBlockRMS])
def test_scalers(constructor: Callable[[], GradientTransformation[Any, Any]]) -> None:
    params = init_params

    scaler = constructor()
    init_fn = variant(scaler.init)
    transform_fn = variant(scaler.update)

    state = init_fn(params)
    chex.assert_tree_all_finite(state)

    updates, state = transform_fn(per_step_updates, state, params)
    chex.assert_tree_all_finite((params, updates, state))
    tree.map(lambda *args: chex.assert_equal_shape(args), params, updates)


def test_add_decayed_weights() -> None:
    # Define a transform that add decayed weights.
    # We can define a mask either as a pytree, or as a function that
    # returns the pytree. Below we define the pytree directly.
    mask = (True, {'a': True, 'b': False})
    tx = AddDecayedWeights[Any](0.1, mask=mask)
    # Define input updates and weights.
    updates = (
        jnp.zeros((2,), dtype=jnp.float32),
        {'a': jnp.zeros((2,), dtype=jnp.float32),
         'b': jnp.zeros((2,), dtype=jnp.float32)})
    weights = (
        jnp.ones((2,), dtype=jnp.float32),
        {'a': jnp.ones((2,), dtype=jnp.float32),
         'b': jnp.ones((2,), dtype=jnp.float32)})
    # This mask means that we will add decayed weights to the first two
    # terms in the input updates, but not to the last element.
    expected_tx_updates = (
        0.1 * jnp.ones((2,), dtype=jnp.float32),
        {'a': 0.1 * jnp.ones((2,), dtype=jnp.float32),
         'b': jnp.zeros((2,), dtype=jnp.float32)})
    # Apply transform
    state = tx.init(weights)
    transform_fn = variant(tx.update)
    new_updates, _ = transform_fn(updates, state, weights)
    # Assert output as expected.
    chex.assert_trees_all_close(new_updates, expected_tx_updates)


def test_ema() -> None:
    values = jnp.array([5.0, 7.0])
    decay = 0.9
    d = decay

    ema = Ema[Any](decay=decay, debias=False)
    state = ema.init(values[0])  # init to zeroes

    transform_fn = variant(ema.update)
    mean, state = transform_fn(values[0], state)
    np.testing.assert_allclose(mean, (1 - d) * values[0], atol=1e-4)

    mean, state = transform_fn(values[1], state)
    np.testing.assert_allclose(mean, (1 - d) * (values[1] + d * values[0]), atol=1e-2)


def test_ema_debias() -> None:
    values = jnp.array([5.0, 7.0])
    decay = 0.9
    d = decay

    ema = Ema[Any](decay=decay)
    state = ema.init(values[0])

    transform_fn = variant(ema.update)
    mean, state = transform_fn(values[0], state)
    np.testing.assert_allclose(mean, values[0], atol=1e-4)

    mean, state = transform_fn(values[1], state)
    np.testing.assert_allclose(mean, ((1 - d) * values[1] + d * (1 - d) * values[0]) / (1 - d**2),
                               atol=1e-2)
    # The state must not be debiased.
    np.testing.assert_allclose(state.data.ema,
                               (1 - d) * values[1] + d * (1 - d) * values[0],
                               atol=1e-2)


def test_apply_every() -> None:
    # The frequency of the application of SGD
    k = 4
    zero_update = (jnp.array([0., 0.]), jnp.array([0., 0.]))

    # optax SGD
    optax_sgd_params = init_params
    sgd_transform = SGD[tuple[RealArray, RealArray]](LR, 0.0)
    state_sgd = sgd_transform.init(optax_sgd_params)

    # optax SGD plus apply every
    optax_sgd_apply_every_params = init_params
    sgd_apply_every = ChainedGradientTransformation[tuple[RealArray, RealArray]](
        [ApplyEvery[Any](k=k),
         Trace[Any](decay=0, nesterov=False),
         Scale[Any](-LR)])
    state_sgd_apply_every = sgd_apply_every.init(optax_sgd_apply_every_params)
    transform_fn = variant(sgd_apply_every.update)

    for i in range(STEPS):
        # Apply a step of SGD
        updates_sgd, state_sgd = sgd_transform.update(per_step_updates, state_sgd, None)
        optax_sgd_params = apply_updates(optax_sgd_params, updates_sgd)

        # Apply a step of sgd_apply_every
        updates_sgd_apply_every, state_sgd_apply_every = transform_fn(
            per_step_updates, state_sgd_apply_every, None)
        optax_sgd_apply_every_params = apply_updates(
            optax_sgd_apply_every_params, updates_sgd_apply_every)

        # Every k steps, check equivalence.
        if i % k == k - 1:
            chex.assert_trees_all_close(
                optax_sgd_apply_every_params, optax_sgd_params,
                atol=1e-6, rtol=1e-5)
        # Otherwise, check update is zero.
        else:
            chex.assert_trees_all_close(
                updates_sgd_apply_every, zero_update, atol=0.0, rtol=0.0)


def test_scale() -> None:
    updates = per_step_updates
    for i in range(1, STEPS + 1):
        factor = 0.1 ** i
        rescaler = Scale[Any](factor)
        empty_state = rescaler.init(updates)
        # Apply rescaling.
        scaled_updates, _ = rescaler.update(updates, empty_state, updates)

        # Manually scale updates.
        def rescale(t: Any, *, factor: float = factor) -> Any:
            return t * factor
        manual_updates = tree.map(rescale, updates)
        # Check the rescaled updates match.
        chex.assert_trees_all_close(scaled_updates, manual_updates)


@pytest.mark.parametrize(("inputs", "outputs"), [
    ([1.0, 2.0], [1.0, 2.0]),
    ([[1.0, 2.0], [3.0, 4.0]], [[-0.5, 0.5], [-0.5, 0.5]]),
    ([[[1., 2.], [3., 4.]],
      [[5., 6.], [7., 8.]]], [[[-1.5, -0.5], [0.5, 1.5]], [[-1.5, -0.5], [0.5, 1.5]]]),
])
def test_centralize(inputs: Any, outputs: Any) -> None:
    inputs = jnp.asarray(inputs)
    outputs = jnp.asarray(outputs)
    centralizer = Centralize[Any]()
    empty_state = centralizer.init(inputs)
    centralized_inputs, _ = centralizer.update(inputs, empty_state, None)
    chex.assert_trees_all_close(centralized_inputs, outputs)



from __future__ import annotations

from collections.abc import Sequence
from types import ModuleType
from typing import Any

import numpy as np
import numpy.typing as npt
from jax import Array as JaxArray

Shape = tuple[int, ...]
ShapeLike = int | Sequence[int]
_SliceLikeItem = int | slice | None
SliceLike = _SliceLikeItem | tuple[_SliceLikeItem, ...]


JaxBooleanArray = JaxArray
JaxIntegralArray = JaxArray
JaxRealArray = JaxArray
JaxComplexArray = JaxArray
KeyArray = JaxArray


NumpyArray = npt.NDArray[Any]
NumpyBooleanArray = npt.NDArray[np.bool_]
NumpyIntegralArray = npt.NDArray[np.integer[Any]]
NumpyRealArray = npt.NDArray[np.floating[Any]]
NumpyComplexArray = npt.NDArray[np.floating[Any]] | npt.NDArray[np.complexfloating[Any, Any]]
BooleanArray = NumpyBooleanArray | JaxArray
IntegralArray = NumpyIntegralArray | JaxArray
RealArray = NumpyRealArray | JaxArray
ComplexArray = NumpyComplexArray | JaxArray


NumpyBooleanNumeric = NumpyBooleanArray | bool
NumpyIntegralNumeric = NumpyIntegralArray | int
NumpyRealNumeric = NumpyRealArray | float
NumpyComplexNumeric = NumpyComplexArray | complex
BooleanNumeric = BooleanArray | bool
IntegralNumeric = IntegralArray | int
RealNumeric = RealArray | float
ComplexNumeric = ComplexArray | complex


# Eventually, these will come from array_api_extra.
Array = NumpyArray | JaxArray
Device = Any
DType = Any
Namespace = ModuleType


PyTree = Any



from __future__ import annotations

from typing import Literal, TypeVar

from array_api_compat import array_namespace, is_jax_array, is_torch_array

from .annotations import Array, BooleanArray, Namespace

T = TypeVar('T', bound=Array)


def abs_square(x: T) -> T:
    xp = array_namespace(x)
    return xp.square(x.real) + xp.square(x.imag)  # pyright: ignore


# TODO: Remove when it's added to theArrayAPI:
# https://github.com/data-apis/array-api/issues/242
def outer_product(x: T, y: T) -> T:
    """Return the broadcasted outer product of a vector with itself.

    This is xp.einsum("...i,...j->...ij", x, y).
    """
    xp = array_namespace(x, y)
    xi = xp.reshape(x, (*x.shape, 1))
    yj = xp.reshape(y.conjugate(), (*y.shape[:-1], 1, y.shape[-1]))
    return xi * yj


def matrix_vector_mul(x: T, y: T) -> T:
    """Return the matrix-vector product.

    This is xp.einsum("...ij,...j->...i", x, y).

    Note the speed difference:
    * 14.3 µs: xp.vecdot(matrix_vector_mul(m, x), x)
    * 4.44 µs: np.einsum("...i,...ij,...j->...", x, m, x)
    """  # noqa: RUF002
    xp = array_namespace(x, y)
    y = xp.reshape(y, (*y.shape[:-1], 1, y.shape[-1]))
    return xp.sum(x * y, axis=-1)


def matrix_dot_product(x: T, y: T) -> T:
    """Return the "matrix dot product" of a matrix with the outer product of a vector.

    This equals:
    * 1.19 µs: xp.einsum("...ij,...ij", x, y)
    * 1.77 µs: xp.sum(x * y, axis=(-2, -1))
    # 3.87 µs: xp.linalg.trace(xp.moveaxis(x, -2, -1) @ y)
    """  # noqa: RUF002
    xp = array_namespace(x, y)
    return xp.sum(x * y, axis=(-2, -1))


def divide_where(dividend: T,
                 divisor: T,
                 *,
                 where: BooleanArray | None = None,
                 otherwise: T | None = None) -> T:
    """Return the quotient or a special value when a condition is false.

    Returns: `xp.where(where, dividend / divisor, otherwise)`, but without evaluating
    `dividend / divisor` when `where` is false.  This prevents both infinite cotangents, and
    some exceptions.
    """
    if where is None:
        assert otherwise is None
        xp = array_namespace(dividend, divisor)
        return xp.divide(dividend, divisor)
    assert otherwise is not None
    xp = array_namespace(dividend, divisor, where, otherwise)
    dividend = xp.where(where, dividend, 1.0)
    divisor = xp.where(where, divisor, 1.0)
    quotient: T = xp.divide(dividend, divisor)
    return xp.where(where, quotient, otherwise)


# Remove when https://github.com/scipy/scipy/pull/18605 is released.
def softplus(x: T, /, *, xp: Namespace | None = None) -> T:
    """Softplus, which is log(1+exp(x)).

    This has asymptotic behavior exp(x) as x -> -inf and x as x -> +inf.
    """
    if xp is None:
        xp = array_namespace(x)
    return xp.logaddexp(xp.asarray(0.0), x)


def log_softplus(x: T, /, *, xp: Namespace | None = None) -> T:
    """Log-softplus, which is log(1+log(1+exp(x))).

    This has asymptotic behavior 0 as x -> -inf and log(x) as x -> +inf.
    """
    if xp is None:
        xp = array_namespace(x)
    z = xp.asarray(0.0)
    return xp.logaddexp(z, xp.logaddexp(z, x))


def sublinear_softplus(x: T, maximum: T, /, *, xp: Namespace | None = None) -> T:
    """Sublinear-softplus, which is softplus(x) / (1 + softplus(x) / maximum).

    This has asymptotic behavior exp(x) as x -> -inf and maximum as x -> +inf.
    """
    if xp is None:
        xp = array_namespace(x)
    o = xp.asarray(1.0)
    sp = softplus(x, xp=xp)
    return sp / (o + sp / maximum)


def inverse_softplus(y: T, /, *, xp: Namespace | None = None) -> T:
    if xp is None:
        xp = array_namespace(y)
    return xp.where(y > 80.0,  # noqa: PLR2004
                    y,
                    xp.log(xp.expm1(y)))


def normalize(mode: Literal['l1', 'l2', 'max'],
              x: T,
              *,
              axis: tuple[int, ...] | int | None = None
              ) -> T:
    """Returns the L1-normalized copy of x, assuming that x is nonnegative."""
    xp = array_namespace(x)
    epsilon = 10 * xp.finfo(x.dtype).eps
    match mode:
        case 'l1':
            sum_x = xp.sum(xp.abs(x), axis=axis, keepdims=True)
            size = x.size / sum_x.size
            return xp.where(sum_x < epsilon, xp.ones_like(x) / size, x / sum_x)
        case 'l2':
            sum_x = xp.sqrt(xp.sum(xp.square(x), axis=axis, keepdims=True))
            size = x.size / sum_x.size
            return xp.where(sum_x < epsilon, xp.ones_like(x) * xp.pow(size, -0.5), x / sum_x)
        case 'max':
            sum_x = xp.max(xp.abs(x), axis=axis, keepdims=True)
            return xp.where(sum_x < epsilon, xp.ones_like(x), x / sum_x)


U = TypeVar('U')


def stop_gradient(x: U, *, xp: Namespace | None = None) -> U:
    if xp is None:
        xp = array_namespace(x)
    if is_jax_array(xp):
        from jax.lax import stop_gradient as sg  # noqa: PLC0415
        return sg(x)
    if is_torch_array(xp):
        from torch import Tensor  # noqa: PLC0415
        assert isinstance(x, Tensor)
        return x.detach()  # pyright: ignore
    return x



import jax.numpy as jnp
import numpy as np
from jax import Array
from numpy.testing import assert_allclose

from tjax import JaxRealArray, leaky_integrate


def test_time_step_invariance() -> None:
    def f(n: int) -> JaxRealArray:
        x = jnp.asarray(0.0)
        for _ in range(n):
            x = leaky_integrate(x,
                                time_step=jnp.asarray(1 / n),
                                drift=jnp.asarray(3.0),
                                decay=jnp.asarray(2.0))
        assert isinstance(x, Array)
        return x

    a = np.asarray([f(n) for n in (1, 5, 20, 100)],
                   dtype=np.float64)
    assert_allclose(a, a[0])
